import json
import logging
from inspect import stack

import click
from click import BadParameter

from com.enovation.helper.click.python_literal_argument_and_option import PythonLiteralArgument

_logger: logging.Logger = logging.getLogger(__name__)


@click.command('dict-load-json')
@click.pass_context
@click.argument('file', type=click.Path(exists=True))
@click.argument('alias', type=str, default='json')
def dict_load_json(ctx_context, file, alias):
    """
    Load json file into a dictionary that is labelled using an alias for later use.

    :param alias: name given to the dictionary being loaded, so it can be used later on by other commands.
    :param ctx_context: Click context.
    :param file: json file to load
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias}' already exists, of type {type(alias)}. This data "
            f"will be overridden with the 'json' file '{file}'."
        )

    # We load the json
    with open(file, 'r') as json_file:
        _dict_the_file: dict = json.load(json_file)

    _logger.info(f"json file read as a dictionary containing '{len(_dict_the_file)}' records.")

    # We update the context data store
    ctx_context.obj[alias] = _dict_the_file
    ctx_context.obj["_" + alias] = {
        "path": file,
        "src": "dict_load_json"
    }

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


@click.command('dict-set-json')
@click.pass_context
@click.argument('dictionary', cls=PythonLiteralArgument, type=dict, default={})
@click.argument('alias', type=str, default={})
def dict_set_json(ctx_context, dictionary, alias):
    """
    Set a DICTIONARY provided as  as an argument as an ALIAS that can be later accessed from other commands
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias in ctx_context.obj:
        _logger.warning(
            f"Warning: another object with alias '{alias}' already exists, of type {type(alias)}. This data "
            f"will be overridden with the 'dictionary' value '{dictionary}'."
        )

    # We update the context data store
    ctx_context.obj[alias] = dictionary
    ctx_context.obj["_" + alias] = {
        "src": "dict_set_json"
    }

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


@click.command('dict-modif')
@click.pass_context
@click.argument('alias', type=str)
@click.argument('delta', cls=PythonLiteralArgument, type=dict, default={})
def dict_modif(ctx_context, alias, delta):
    """
    Modify the ALIAS dictionary from the context using a DELTA dictionary.
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if alias not in ctx_context.obj:
        raise BadParameter(
            message=f"No object with alias '{alias}' exists in the context. Cannot do any magic here!",
            param_hint=f"Ensure you effectively loaded a dictionary in the context, with alias '{alias}', before "
                       f"trying to modify it.",
            ctx=ctx_context
        )

    _alias_object = ctx_context.obj[alias]

    if not isinstance(_alias_object, dict):
        raise BadParameter(
            message=f"Object with alias '{alias}' in context is of type '{type(_alias_object)}', while 'dict' is "
                    f"expected.",
            param_hint=f"Ensure you effectively try to modify an object in the context which is of type 'dict'.",
            ctx=ctx_context
        )

    _alias_dict: dict = _alias_object

    _alias_dict.update(delta)

    # We update the context data store
    ctx_context.obj[alias] = _alias_dict
    ctx_context.obj["_" + alias] = {
        "src": "dict_modif"
    }

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from inspect import stack
from logging import Logger, getLogger
from typing import Tuple

import numpy as np
from jsonschema import validate
from numpy import array
from pandas import DataFrame

from com.enovation.helper.pandas_dataframe_typer import PandasDataframeTyper


class BacklogComparer:
    _logger: Logger = getLogger(__name__)

    str__source_column_suffix: str = "_from"
    str__target_column_suffix: str = "_to"
    str__delta_column_suffix: str = "_delta"

    str__dict_columns__json_tag__key_columns: str = "key columns"
    str__dict_columns__json_tag__compared_columns: str = "compared columns"
    str__dict_columns__json_tag__type: str = "type"
    str__dict_columns__json_tag__compare_logic: str = "compare logic"
    str__dict_columns__json_tag__delta_type: str = "delta type"

    lst__dict_columns__json_tags: list[str] = [
        str__dict_columns__json_tag__key_columns,
        str__dict_columns__json_tag__compared_columns,
        str__dict_columns__json_tag__type
    ]

    def compare(
            self,
            df_src: DataFrame,
            df_tgt: DataFrame,
            dict_columns: dict
    ) -> DataFrame:
        """

        :param df_src:
        :param df_tgt:
        :param dict_columns: dictionary
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # From the dict_columns, we prepare the various sub-dictionaries/ lists
        dict_the_columns_types: dict[str]  # A dictionary to properly type the dataframes (both key and comparable col)
        lst_the_compared_columns: list[str]  # A list of the columns' labels to compare
        lst_the_key_columns: list[str]  # A list of the key columns' labels to merge dataframes
        dict_the_columns_comparison_logics: dict  # A dictionary to properly associate comparison functions to columns

        # First, we check the dict_columns. Exception will be thrown in case anything is incorrect
        self._check_parameter_dict_columns(
            df_src=df_src,
            df_tgt=df_tgt,
            dict_columns=dict_columns
        )

        # First, we process the dict_columns, and raise an exception should its format be incorrect
        dict_the_columns_types, lst_the_compared_columns, lst_the_key_columns, dict_the_columns_comparison_logics = \
            self._process_dict_columns(
                dict_columns=dict_columns
            )

        # We type and merge columns of interest from both dataframes into one single dataframe
        df_the_merged_dataframes: DataFrame = self._type_and_merge_data_frames(
            df_src=df_src,
            df_tgt=df_tgt,
            dict_columns_types=dict_the_columns_types,
            lst_compared_columns=lst_the_compared_columns,
            lst_key_columns=lst_the_key_columns
        )

        df_the_delta: DataFrame = self._reconcile(
            df=df_the_merged_dataframes,
            dict_columns_comparison_logics=dict_the_columns_comparison_logics
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return df_the_delta

    def compare_into_excel(
            self,
            df_src: DataFrame,
            df_tgt: DataFrame,
            dict_columns: dict,
            str_path: str
    ):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        df_the_results: DataFrame = self.compare(
            df_src=df_src,
            df_tgt=df_tgt,
            dict_columns=dict_columns
        )

        df_the_results.to_excel(str_path)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def _check_parameter_dict_columns(
            self,
            df_src: DataFrame,
            df_tgt: DataFrame,
            dict_columns: dict[str]
    ):
        """
        Function that check the parameter dict_columns provided as an input:
            - If parameter is correct, the function is returning (no value returned)
            - Otherwise, an exception is thrown.

        dict_columns should follow the following structure:
          {
            "key columns": {
              "the label of the key column": {
                "type": "the type of the key column, as defined in PandasDataframeTyper.lst__types",
              },
            },
            "compared columns": {
              "the label of the column to compare": {
                "type": "the type of the column to compare, as defined in PandasDataframeTyper.lst__types",
                "compare logic": "the function to compare",
                "delta type": "the type of the delta, as defined in PandasDataframeTyper.lst__types"
              }
            }
          }

        ---

        Business Rules:
            - BR_101: parameter "key columns" should exist
            - BR_102: parameter "key columns" should be a dictionary
            - BR_103: parameter "key columns" should contain at least one property
            - BR 104: parameter "key columns" should only contains keys among columns intersecting in both dataframes
            - BR_105: parameter "key columns" should contains records containing "type" key only
            - BR_106: parameter "key columns" should contains records containing "type" key
            - BR_107: parameter "key columns - type" should be among the ones defined in
              PandasDataframeTyper.lst__types

            - BR_201: parameter "compared columns" should exist
            - BR_202: parameter "compared columns" should be a dictionary
            - BR_203: parameter "compared columns" should contain at least one property
            - BR 204: parameter "compared columns" should only contains keys among columns intersecting in both dfs
            - BR_205: parameter "compared columns" should contains records containing "type" key only
            - BR_206: parameter "compared columns" should contains records containing "type" key
            - BR_207: parameter "compared columns - type" should be among the ones defined in
              PandasDataframeTyper.lst__types

            - BR_301: "key columns" and "compared columns" should be distinct. Aka no column should be listed as
              a key and a compared column at the same time

        :param df_tgt:
        :param df_src:
        :param dict_columns:
        :return: a tuple made of
          - dict_columns_types: the dictionary to type the dataframe (see PandasDataframeTyper)
          - lst_compared_columns: the list of columns to compare
          - lst_key_columns: the list of columns to be used as key to compare records across dataframes
        """

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # First, we get the union of columns across both dataframes
        array_the_columns: array = np.intersect1d(df_src.columns, df_tgt.columns)

        dict_the_json_schema: dict = {

            "title": "id1",

            # Type dictionary
            "type": "object",

            # with properties
            "properties": {

                # "key columns"
                self.str__dict_columns__json_tag__key_columns: {

                    # of type dictionary
                    "type": "object",

                    # that contains at least one record
                    "minProperties": 1,

                    # among the following properties
                    "properties": {

                        # that could be any column in common across the 2 dataframes
                        i_col: {"$ref": "#/$defs/key_column"}
                        for i_col in array_the_columns
                    },

                    # we forbid any other property
                    "additionalProperties": False
                },

                # The second node is the compared columns
                self.str__dict_columns__json_tag__compared_columns: {

                    # of type dictionary
                    "type": "object",

                    # that contains at least one record
                    "minProperties": 1,

                    # among the following properties
                    "properties": {

                        # that could be any column in common across the 2 dataframes
                        i_col: {"$ref": "#/$defs/compared_column"}
                        for i_col in array_the_columns
                    },

                    # we forbid any other property
                    "additionalProperties": False
                },
            },

            # that requires properties
            "required": [
                self.str__dict_columns__json_tag__key_columns,
                self.str__dict_columns__json_tag__compared_columns
            ],

            "$defs": {

                # Sub-type "key_column"
                "key_column": {

                    # Type dictionary
                    "type": "object",

                    # with properties
                    "properties": {
                        # "type"
                        self.str__dict_columns__json_tag__type: {
                            # of type string
                            "type": "string",
                            # among predefined values
                            "enum": PandasDataframeTyper.lst__types
                        }
                    },

                    # we forbid any other property
                    "additionalProperties": False,

                    # that requires properties
                    "required": [
                        self.str__dict_columns__json_tag__type
                    ],
                },

                # Sub-type "compared_column"
                "compared_column": {

                    # Type dictionary
                    "type": "object",

                    # with properties
                    "properties": {
                        # "type"
                        self.str__dict_columns__json_tag__type: {
                            # of type string
                            "type": "string",
                            # among predefined values
                            "enum": PandasDataframeTyper.lst__types
                        },
                        # compare logic
                        self.str__dict_columns__json_tag__compare_logic: {
                            # of type string
                            "type": "string",
                        },
                        # delta type
                        self.str__dict_columns__json_tag__delta_type: {
                            # of type string
                            "type": "string",
                        }

                    },

                    # we forbid any other property
                    "additionalProperties": False,

                    # that requires properties
                    "required": [
                        self.str__dict_columns__json_tag__type,
                        # self.str__dict_columns__json_tag__compare_logic -> If not set, we test equality using "=="
                        # self.str__dict_columns__json_tag__delta_type -> If not set, we use the column type
                    ],
                }
            },
        }

        self._logger.debug(f"_check_parameter_dict_columns relies on a jsonschema which is:\n{dict_the_json_schema}")

        # Additional check: BR_301:
        # "key columns" and "compared columns" should be distinct. Aka no column should be listed as a key and a
        # compared column at the same time
        lst_intersecting_columns: list[str] = [
            i_key
            for i_key in (
                    dict_columns[self.str__dict_columns__json_tag__key_columns].keys()
                    & dict_columns[self.str__dict_columns__json_tag__compared_columns].keys()
            )
        ]

        if len(lst_intersecting_columns) > 0:
            raise Exception(
                f"the parameter dict_columns contains columns defined as both a key and a compared column, while these "
                f"universes should not intersect.\n"
                f"\t Faulty columns: {','.join(lst_intersecting_columns)}"
            )

        validate(
            instance=dict_columns,
            schema=dict_the_json_schema
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def _process_dict_columns(
            self,
            dict_columns: dict[str]
    ) -> Tuple[
        dict[str],  # dict_columns_types: the dictionary to properly type the dataframes (both key and comparable col)
        list[str],  # the list of the columns' labels to compare
        list[str],  # the list of the key columns' labels to merge dataframes
        dict        # the dictionary to properly associate comparison functions to columns
    ]:
        """
        Function that process the dict_columns provided as an input, and return the various lists and dictionaries
        later used in the logic

        Assumption: dict_columns were previously checked (through function _check_parameter_dict_columns

        :param dict_columns:
        :return: a tuple made of
          - dict_columns_types: the dictionary to properly type the dataframes (both key and comparable col)
            (see PandasDataframeTyper)
          - lst_compared_columns: the list of the columns' labels to compare
          - lst_key_columns: the list of the key columns' labels to merge dataframes
          - dict_columns_comparison_logics: the dictionary to properly associate comparison functions to columns
        """

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # #############################################################################################################
        # 1. We prepare the dictionary used to type the dataframes
        dict_return_columns_types: dict[str] = {
            i_key: i_val[self.str__dict_columns__json_tag__type]
            for i_key, i_val in dict_columns[self.str__dict_columns__json_tag__key_columns].items()
        }
        dict_return_columns_types.update(
            {
                i_key: i_val[self.str__dict_columns__json_tag__type]
                for i_key, i_val in dict_columns[self.str__dict_columns__json_tag__compared_columns].items()
            }
        )

        # #############################################################################################################
        # 2. We prepare the list of columns to compare. Other columns will be discarded
        lst_return_comparable_columns: list[str] = [
            i_key
            for i_key in dict_columns[self.str__dict_columns__json_tag__compared_columns]
        ]

        # #############################################################################################################
        # 3. We prepare the list of key columns to merge dataframes.
        lst_return_key_columns: list[str] = [
            i_key
            for i_key in dict_columns[self.str__dict_columns__json_tag__key_columns]
        ]

        # #############################################################################################################
        # 4. We prepare the dictionary with the comparison logics as functions
        dict_columns_comparison_logics: dict = {
            i_key:
                f"("
                f"  ("
                f"    row['{i_key + self.str__source_column_suffix}'].isnull() "
                f"    & row['{i_key + self.str__target_column_suffix}'].isnull()"
                f"  )"
                f"  | ("
                f"    row['{i_key + self.str__source_column_suffix}'] "
                f"    == row['{i_key + self.str__target_column_suffix}']"
                f"  )"
                f")"
                if self.str__dict_columns__json_tag__compare_logic not in i_value
                else
                f"("
                f"  ("
                f"    row['{i_key + self.str__source_column_suffix}'].isnull() "
                f"    & row['{i_key + self.str__target_column_suffix}'].isnull()"
                f"  )"
                f"  | ("
                + i_value[self.str__dict_columns__json_tag__compare_logic].replace(
                        "src",
                        f"row['{i_key + self.str__source_column_suffix}']"
                ).replace(
                    "tgt",
                    f"row['{i_key + self.str__target_column_suffix}']"
                ) +
                f"  )"
                f")"

            for i_key, i_value in dict_columns[self.str__dict_columns__json_tag__compared_columns].items()
        }

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

        return \
            dict_return_columns_types, \
            lst_return_comparable_columns, \
            lst_return_key_columns, \
            dict_columns_comparison_logics

    def _type_and_merge_data_frames(
            self,
            df_src: DataFrame,
            df_tgt: DataFrame,
            dict_columns_types: dict,
            lst_key_columns: list[str],
            lst_compared_columns: list[str],
    ) -> DataFrame:
        """
        Dataframes received as inputs comply to the following:
        - Columns are properly typed
        - Only columns to compare are there.

        :param df_src:
        :param df_tgt:
        :param lst_key_columns:
        :param lst_compared_columns:
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We filter out the useless columns
        df_the_filtered_src: DataFrame = df_src[lst_key_columns + lst_compared_columns]
        df_the_filtered_tgt: DataFrame = df_tgt[lst_key_columns + lst_compared_columns]

        # We type the columns
        df_the_typed_src: DataFrame = PandasDataframeTyper().type(
            df_to_type=df_the_filtered_src.copy(),
            dict_columns_to_type=dict_columns_types,
            b_strict=True
        )
        df_the_typed_tgt: DataFrame = PandasDataframeTyper().type(
            df_to_type=df_the_filtered_tgt.copy(),
            dict_columns_to_type=dict_columns_types,
            b_strict=True
        )

        # We merge
        df_the_return: DataFrame = df_the_typed_tgt.merge(
            right=df_the_typed_src,
            on=lst_key_columns,
            suffixes=(
                self.str__target_column_suffix,
                self.str__source_column_suffix,
            )
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")
        return df_the_return

    def _reconcile(
            self,
            df: DataFrame,
            dict_columns_comparison_logics: dict
    ) -> DataFrame:
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        df_the_return: DataFrame = df

        for i_column_label, i_logic in dict_columns_comparison_logics.items():
            df_the_return = df_the_return.assign(
                **{
                    i_column_label + self.str__delta_column_suffix: lambda row: eval(i_logic)
                }
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")
        return df_the_return



from inspect import stack
from logging import Logger, getLogger
from typing import Optional, Union

from xlsxwriter.format import Format
from xlsxwriter.utility import xl_cell_to_rowcol, xl_rowcol_to_cell
from xlsxwriter.workbook import Workbook
from xlsxwriter.worksheet import Worksheet

from com.enovation.toolbox.excel_dashboard.cell_notation_helper import CellNotationHelper
from com.enovation.toolbox.excel_dashboard.widgets.widget import Widget


class WidgetTableZoom(Widget):
    """
    Widget that prints detailed characteristics for a given key from a table.

    Configuration:
    - The cell with the key to select the record to zoom from the table which contains all the details
    - The list of characteristics to display the one after the other, along with the following details:
      - The formula to get and display the values
      - The format
      - The conditional formatting.

    Parameters:
    - No parameter is provided.
    """

    # The labels to parse the json configuration file
    const_str__config_label__str_parameter: str = "parameter"
    const_str__config_label__str_value: str = "value"
    const_str__config_label__str_format: str = "format"
    const_str__config_label__str_conditional_formattings: str = "conditional_formattings"
    const_str__config_label__str_merged: str = "merged"

    const_str__function_label__str_merge_range: str = "merge_range"
    const_str__function_label__str_write: str = "write"
    const_str__function_label__str_function: str = "function"
    const_str__function_label__str_parameter: str = "parameter"

    # The labels for the parameters expected by the widget
    # Illustration: const_str__parameter_label__df_data: DataFrame = "df_data"

    _logger: Logger = getLogger(__name__)

    @property
    def dict_cells_or_ranges_to_write(self) -> dict:
        return self._dict_cells_or_ranges_to_write

    def __init__(
            self,
            str_address: str,
            dict_parameters_map: dict,
            dict_config: dict,
            dict_raw_parameters: dict,
            str_sheet_name: str,
            dict_default_format: dict
    ):
        """
        Initialize the widget:
        - Calling the Widget super init function
        - Declare the widget properties:
          - Each has their "property" getters
          - They are declared and instantiated to None
          - They are effectively set through the function "check_parameters_and_set_properties"

        :param str_address: the address in a sheet to print the widget
        :param dict_config: the widget configuration (aka the "config" node in the json)
        :param dict_parameters_map: the dictionary to map users parameters to the ones expected by the widget:
                                    - key: the expected label
                                    - value: the actual label
        :param dict_raw_parameters: the parameters as labelled and provided by the users
        :param str_sheet_name: the name of the sheet in which the widget is instantiated (required to handle default
            formats inheritance, from the workbook to the workshoeet to the widget)
        :param dict_default_format: the default format set at the level of the widget. If such default format is not
            configured, we expect an empty dictionary
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We declare the widget properties, initialized to None
        # Illustration: self._dict_options: Optional[dict] = None
        self._dict_cells_or_ranges_to_write: Optional[dict] = None
        # self._dict_options: Optional[dict] = None

        # The super init function initializes the widget:
        # - Step 1, we set the widget properties (from nodes from the json configuration): address, parameters map,
        #   config
        # - Step 2, we process the parameters:
        #   - Rename parameters to fit the labels as expected by the widget
        #   - Check parameters to ensure their values are correct, and set widget properties, to use these parameters
        #     while printing into the Excel spreadsheet --> function 'check_parameters_and_set_properties' to be
        #     overridden by each widget will call both the functions to check parameters and config
        super().__init__(
            str_address=str_address,
            dict_parameters_map=dict_parameters_map,
            dict_config=dict_config,
            dict_raw_parameters=dict_raw_parameters,
            str_sheet_name=str_sheet_name,
            dict_default_format=dict_default_format
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def check_parameters_and_set_properties(
            self,
            dict_parameters: dict,
            dict_config: dict,
    ):
        """
        Function that checks the parameters that are required by the widget:
        - If all expected parameters are correct, the widget properties are set and the function returns (no value
          returned)
        - Otherwise, an exception is thrown.

        Checking parameters means:
        - Check the parameter exists, in case it is a mandatory one. eg. mandatory array of data to print in a list
        - Check the format is as expected. eg. has to be a DataFrame
        - Check the values are correct. eg. numeric vs alpha, etc.
        - Check coherence with config. eg. columns configured as "to be printed" effectively exist in the DataFrame.

        Setting properties means:
        - Set the widget instance properties from the parameters/ config
        - So these properties can later be used while writing into Excel.

        Properties for widget_cover_sheet are:
        - str_narrative
        - lst_tokens

        :param dict_parameters: the parameters provided to the module
        :param dict_config: the configuration of the module
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # #####################################################@
        # We check the parameters

        # We screen the config, and append in a list the parameters to print
        _dict_the_parameters = self._get_parameters(
            config=dict_config,
            dict_parameters=dict_parameters
        )

        # We get the default format/ value/ parameter
        _dict_the_default_format: dict = dict_config.get(self.const_str__config_label__str_format, {})
        _str_the_default_parameter: str = dict_config.get(self.const_str__config_label__str_parameter, None)

        if _str_the_default_parameter is not None:
            _str_the_default_value: str = dict_parameters[_str_the_default_parameter]
        else:
            _str_the_default_value: str = dict_config.get(self.const_str__config_label__str_value, "")

        # We then process each area one by one

        self._dict_cells_or_ranges_to_write = {}

        for k, v in dict_config.items():
            # We ignore the keys "format", "value", "parameter" and "conditional formatting"
            if k in [
                self.const_str__config_label__str_format,
                self.const_str__config_label__str_value,
                self.const_str__config_label__str_parameter,
                self.const_str__config_label__str_conditional_formattings
            ]:
                pass

            else:
                self._dict_cells_or_ranges_to_write.update(
                    self._process_range(
                        str_reference_address=self.str_address,
                        dict_parameters=_dict_the_parameters,
                        str_address=k,
                        config=v,
                        dict_default_format=_dict_the_default_format,
                        default_value=_str_the_default_value
                    )
                )

        # We set the conditional formattings
        self.set_conditional_formattings(
            lst_conditional_formattings=dict_config.get(
                self.const_str__config_label__str_conditional_formattings, []
            ),
            str_reference_address=self.str_address,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    @staticmethod
    def _process_range(
            str_reference_address: str,
            str_address: str,
            config: Union[dict, list],
            dict_parameters: dict,
            dict_default_format: dict,
            default_value
    ) -> dict:
        """

        :param str_reference_address:
        :param str_address:
        :param config:
        :param dict_parameters:
        :param dict_default_format: cannot be None, only {}
        :param default_value: cannot be None, only ""
        :return:
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # These are the variables we will set, to eventually return a processed range
        _lst_the_cells_addresses: Optional[list[str]] = None
        _lst_the_formats: Optional[list[dict]] = None
        _lst_the_values: Optional[list] = None
        _lst_the_functions: Optional[list[str]] = None

        # We translate the address
        _str_translated_address: str = CellNotationHelper.translate_address_to_reference(
            str_reference_cell=str_reference_address,
            str_address=str_address
        )

        # We decompose the address into a list of cells:
        # - If address is a range, we get a list with all the cells contained in the range
        # - If address is one cell, we get a list with only this cell
        _lst_the_cells_addresses = WidgetTableZoom._for_each_cell_in_range(_str_translated_address)

        # If the configuration is of type "dictionary"
        if isinstance(config, dict):

            # ##########################################################################################################
            # We set the format
            _the_format_to_return: Union[dict, list] = config.get(
                WidgetTableZoom.const_str__config_label__str_format,
                dict_default_format
            )
            if isinstance(_the_format_to_return, list):
                _lst_the_formats = _the_format_to_return
            elif config.get(WidgetTableZoom.const_str__config_label__str_merged, False):
                _lst_the_formats = [_the_format_to_return]
            else:
                _lst_the_formats = [_the_format_to_return] * len(_lst_the_cells_addresses)

            # ##########################################################################################################
            # We set the value

            # Most simple case: we have straight value(s)
            if WidgetTableZoom.const_str__config_label__str_value in config:
                _the_value_to_return = config[WidgetTableZoom.const_str__config_label__str_value]

            # Else, if we have parameter(s) that we need to transcode:
            # - Either a list of parameters to transcode one by one
            # - Or one single parameter
            elif WidgetTableZoom.const_str__config_label__str_parameter in config:

                _the_value_to_return = config[WidgetTableZoom.const_str__config_label__str_parameter]

                if isinstance(_the_value_to_return, list):
                    _the_new_values_to_return: list = []
                    for i_value in _the_value_to_return:
                        if i_value not in dict_parameters:
                            raise Exception(f"Parameter '{i_value}' is missing.")
                        _the_new_values_to_return.append(dict_parameters[i_value])

                    _the_value_to_return = _the_new_values_to_return

                else:
                    if _the_value_to_return not in dict_parameters:
                        raise Exception(f"Parameter '{_the_value_to_return}' is missing.")
                    _the_value_to_return = dict_parameters[_the_value_to_return]

            # Else, we default the values
            else:
                _the_value_to_return = default_value

            # We now ensure the value is a list of values
            if isinstance(_the_value_to_return, list):
                _lst_the_values = _the_value_to_return
            elif config.get(WidgetTableZoom.const_str__config_label__str_merged, False):
                _lst_the_values = [_the_value_to_return]
            else:
                _lst_the_values = [_the_value_to_return] * len(_lst_the_cells_addresses)

            # We set the function
            if config.get(WidgetTableZoom.const_str__config_label__str_merged, False):
                _lst_the_functions = [WidgetTableZoom.const_str__function_label__str_merge_range]
            else:
                _lst_the_functions = [WidgetTableZoom.const_str__function_label__str_write] *\
                                     len(_lst_the_cells_addresses)

        # Else, the configuration is of type list
        else:

            _lst_the_formats = []
            _lst_the_values = []
            _lst_the_functions = []

            for i_config in config:

                # We set the format
                _dict_the_format_to_return: dict = i_config.get(
                    WidgetTableZoom.const_str__config_label__str_format,
                    dict_default_format
                )
                _lst_the_formats.append(_dict_the_format_to_return)

                # We set the value
                _the_value_to_return = i_config.get(
                    WidgetTableZoom.const_str__config_label__str_parameter,
                    i_config.get(
                        WidgetTableZoom.const_str__config_label__str_value,
                        default_value
                    )
                )
                _lst_the_values.append(_the_value_to_return)

                # We set the function
                _lst_the_functions.append(WidgetTableZoom.const_str__function_label__str_write)

        _int_the_range_size: int = len(_lst_the_cells_addresses)

        # If the area is to be merged
        if _lst_the_functions[0] == WidgetTableZoom.const_str__function_label__str_merge_range:

            # We check we have the proper number of records
            if (len(_lst_the_functions) != 1) \
                    | (len(_lst_the_formats) != 1) \
                    | (len(_lst_the_values) != 1):
                raise Exception(f"For the range '{str_address}' to merge, we have '{_int_the_range_size}' "
                                f"cells to process, but '{len(_lst_the_functions)}' functions, '{len(_lst_the_values)}'"
                                f" values and '{len(_lst_the_formats)}' formats. WE EXPECTED ALL TO BE ONE!")

            # Eventually, we produce the processed range
            _dict_the_return: dict = {
                _str_translated_address.replace("$", ""): {
                    # Function: "merge" or "write"
                    WidgetTableZoom.const_str__function_label__str_function: _lst_the_functions[0],

                    # Value and format
                    WidgetTableZoom.const_str__function_label__str_parameter: {
                        WidgetTableZoom.const_str__config_label__str_format: _lst_the_formats[0],
                        WidgetTableZoom.const_str__config_label__str_value: _lst_the_values[0]
                    }
                }
            }

        # Else, the area is not to be merged
        else:

            # We check we have the proper number of records
            if (len(_lst_the_functions) != _int_the_range_size) \
                    | (len(_lst_the_formats) != _int_the_range_size) \
                    | (len(_lst_the_values) != _int_the_range_size):
                raise Exception(f"For the range '{str_address}', we have '{_int_the_range_size}' cells to process, but "
                                f"'{len(_lst_the_functions)}' functions, '{len(_lst_the_values)}' values and "
                                f"'{len(_lst_the_formats)}' formats. WE EXPECTED ALL TO BE EQUAL TO "
                                f"'{_int_the_range_size}'")

            # Eventually, we produce the processed range
            _dict_the_return: dict = {}
            for i_cell in range(0, len(_lst_the_cells_addresses)):
                _dict_the_return[_lst_the_cells_addresses[i_cell]] = {
                    # Function: "merge" or "write"
                    WidgetTableZoom.const_str__function_label__str_function: _lst_the_functions[i_cell],

                    # Value and format
                    WidgetTableZoom.const_str__function_label__str_parameter: {
                        WidgetTableZoom.const_str__config_label__str_format: _lst_the_formats[i_cell],
                        WidgetTableZoom.const_str__config_label__str_value: _lst_the_values[i_cell]
                    }
                }

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _dict_the_return

    @staticmethod
    def _for_each_cell_in_range(str_range: str) -> list[str]:
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _lst_cells: list[str] = str_range.split(":")

        if len(_lst_cells) == 1:
            _lst_the_return: list[str] = [str_range.replace("$", "")]

        else:

            (_int_first_row, _int_first_col) = xl_cell_to_rowcol(_lst_cells[0])
            (_int_last_row, _int_last_col) = xl_cell_to_rowcol(_lst_cells[1])

            _lst_the_cell_as_row_col: list = [
                (i_row, i_col)
                for i_row in range(_int_first_row, _int_last_row+1) for i_col in range(_int_first_col, _int_last_col+1)
            ]

            _lst_the_return: list[str] = [
                xl_rowcol_to_cell(row=i_row_col[0], col=i_row_col[1])
                for i_row_col in _lst_the_cell_as_row_col
            ]

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _lst_the_return

    @staticmethod
    def _get_parameters(
            config: Union[dict, list],
            dict_parameters: dict = None
    ) -> dict:
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We list the expected parameters
        _lst_the_parameters: list = WidgetTableZoom._list_parameters(config=config)

        # We check these expected parameters effectively exist
        _lst_the_diff: list = list(set(_lst_the_parameters) - set(dict_parameters.keys()))
        if len(_lst_the_diff) > 0:
            raise Exception(f"Configuration expects parameters which are not provided: {', '.join(_lst_the_diff)}")

        _dict_the_return: dict = {
            k: dict_parameters[k] for k in _lst_the_parameters
        }

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _dict_the_return

    @staticmethod
    def _list_parameters(config: Union[dict, list]) -> list:
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _lst_the_return: list = []

        # If we screen a dictionary
        if isinstance(config, dict):

            # We screen each and every (key, value)
            for k, v in config.items():

                # If key is "parameter"
                if k == WidgetTableZoom.const_str__config_label__str_parameter:

                    # If value is a list
                    if isinstance(v, list):
                        _lst_the_return.extend(v)

                    # If value is not a list
                    else:
                        _lst_the_return.append(v)

                # Else, key is not "parameter
                # And if value is itself a dictionary or a list
                elif isinstance(v, (dict, list)):
                    _lst_the_return.extend(WidgetTableZoom._list_parameters(config=v))

        # Else, we screen a list
        else:

            # For each record in the list
            for v in config:

                # if value is itself a dictionary or a list
                if isinstance(v, (dict, list)):
                    _lst_the_return.extend(WidgetTableZoom._list_parameters(config=v))

        # We then remove duplicates
        _lst_the_return = list(dict.fromkeys(_lst_the_return))

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _lst_the_return

    def write_to_excel(
            self,
            wb_workbook: Workbook,
            ws_worksheet: Worksheet,
    ):
        """
        Function that writes the widget into the excel spreadsheet. When called, the widget is fully initialized,
        with all its properties set.

        Note: it is while writing into excel that we transcode formats, from "format config" as dictionary to "format"
        object as expected by xlsxwriter.

        :param ws_worksheet: the worksheet to print the widget
        :param wb_workbook:
        """

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        for k, v in self.dict_cells_or_ranges_to_write.items():

            _str_the_function: str = v[self.const_str__function_label__str_function]

            _obj_the_format: Format = self.get_format(
                dict_format=v[self.const_str__config_label__str_parameter][self.const_str__config_label__str_format],
                wb_workbook=wb_workbook
            )

            if _str_the_function == self.const_str__function_label__str_write:
                ws_worksheet.write(
                    k,
                    v[self.const_str__config_label__str_parameter][self.const_str__config_label__str_value],
                    _obj_the_format
                )

            elif _str_the_function == self.const_str__function_label__str_merge_range:
                ws_worksheet.merge_range(
                    k,
                    v[self.const_str__config_label__str_parameter][self.const_str__config_label__str_value],
                    _obj_the_format
                )

            else:
                raise Exception(f"Function '{_str_the_function}' is unexpected.")

        # We eventually write the conditional formattings
        self.write_conditional_formattings(
            wb_workbook=wb_workbook,
            ws_worksheet=ws_worksheet
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



from inspect import stack

import click
from click import BadParameter
import logging

_logger: logging.Logger = logging.getLogger(__name__)


# noinspection PyUnusedLocal
@click.command('py-exec-file')
@click.pass_context
@click.argument('file', type=click.Path(exists=True))
def py_exec_file(ctx_context, file):
    """
    Load the content of FILE as text, and execute it.

    Raises BadParameter exception in case:
     - The file cannot be read a text
     - The text cannot be executed.

    :param ctx_context:
    :param file: text file to load the python code from
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    try:
        with open(file, 'r') as _io_the_file:
            _lst_the_lines: list[str] = _io_the_file.readlines()
    except Exception as _the_exception:
        raise BadParameter(f"Exception thrown from function '{stack()[0].filename} - {stack()[0].function}'.") \
            from _the_exception

    try:
        exec("\n".join(_lst_the_lines))
    except Exception as _the_exception:
        raise BadParameter(f"Exception thrown from function '{stack()[0].filename} - {stack()[0].function}'.") \
            from _the_exception

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import logging
import inspect

from numpy import nan
from pandas import DataFrame, to_datetime, isnull


class PandasDataframeTyper:

    _logger: logging.Logger = logging.getLogger(__name__)

    str__type__date = 'date'
    str__type__float = 'float'
    str__type__int = 'int'
    str__type__str = 'str'

    lst__types: list[str] = [
        str__type__date,
        str__type__float,
        str__type__int,
        str__type__str
    ]

    @staticmethod
    def type(
            df_to_type: DataFrame,
            dict_columns_to_type: dict,
            b_strict: bool = False,
    ) -> DataFrame:
        """
        dict_columns_to_type = {
            "a column name": "a format"
        }
        With "a format" among:
        - "date"
        - "float"
        - "int"
        - "str"

        :param b_strict: if True, we raise an exception in case we do not have a match between the columns in the
            dataframe and the dictionary
        :param df_to_type:
        :param dict_columns_to_type:
        :return:
        """

        PandasDataframeTyper._logger.debug(
            f"Function '{inspect.stack()[0].filename} - {inspect.stack()[0].function}' is called.")

        the_df_return: DataFrame = df_to_type

        # We check the dataframe columns versus the ones in the dictionary
        the_extra_columns_in_df: list = list(
            set(df_to_type.columns)
            - set([i_column_name for i_column_name in dict_columns_to_type.keys()])
        )

        the_extra_columns_in_dict: list = list(
            set([i_column_name for i_column_name in dict_columns_to_type.keys()])
            - set(df_to_type.columns)
        )

        if (len(the_extra_columns_in_df) + len(the_extra_columns_in_dict)) > 0:
            the_msg: str = f"When requesting to type a dataframe, we discovered the following discrepancies:\n\t" \
                           f"- Columns in dataframe, but missing in  dictionary: '" \
                           f"{', '.join(the_extra_columns_in_df)}'\n\t" \
                           f"- Columns in dictionary, but missing in dataframe: '" \
                           f"{', '.join(the_extra_columns_in_dict)}'."

            if b_strict:
                raise Exception(the_msg)

            else:
                PandasDataframeTyper._logger.info(the_msg)

        # We go through each columns to be retyped
        for k_field, v_df_type in dict_columns_to_type.items():

            # The field to retype might not exist (if b_strict is false)
            if k_field in df_to_type.columns:

                # Depending on the target type
                if v_df_type == PandasDataframeTyper.str__type__date:
                    try:
                        # JSG as of 22-Jan-2023: FutureWarning ---------------------------------------------------------
                        # FutureWarning: In a future version, `df.iloc[:, i] = newvals` will attempt to set the values
                        # inplace instead of always setting a new array. To retain the old behavior, use either
                        # `df[df.columns[i]] = newvals` or, if columns are non-unique, `df.isetitem(i, newvals)`
                        # FROM -----------------------------------------------------------------------------------------
                        # the_df_return.loc[:, k_field] = to_datetime(
                        #     the_df_return.loc[:, k_field], infer_datetime_format=True).dt.normalize()
                        # TO -------------------------------------------------------------------------------------------
                        the_df_return[k_field] = to_datetime(
                             the_df_return[k_field], infer_datetime_format=True).dt.normalize()
                        # END ------------------------------------------------------------------------------------------
                    except Exception as an_exception:
                        raise Exception(an_exception, f"Exception raised when typing column "
                                                      f"'{k_field}' into '{v_df_type}'.")

                elif v_df_type == PandasDataframeTyper.str__type__float:
                    try:
                        # JSG as of 22-Jan-2023: FutureWarning ---------------------------------------------------------
                        # FutureWarning: In a future version, `df.iloc[:, i] = newvals` will attempt to set the values
                        # inplace instead of always setting a new array. To retain the old behavior, use either
                        # `df[df.columns[i]] = newvals` or, if columns are non-unique, `df.isetitem(i, newvals)`
                        # FROM -----------------------------------------------------------------------------------------
                        # the_df_return.loc[:, k_field] = the_df_return[k_field].astype('Float64')
                        # TO -------------------------------------------------------------------------------------------
                        the_df_return[k_field] = the_df_return[k_field].astype('Float64')
                        # END ------------------------------------------------------------------------------------------

                        # Change as of 28-FEB-22, after upgrading packages:
                        # - Replaced below pandas.NaT by numpy.nan
                        # - In order to avoid an unexpected exception while typing to Float64 the column

                        # JSG as of 22-Jan-2023: FutureWarning ---------------------------------------------------------
                        # FutureWarning: In a future version, `df.iloc[:, i] = newvals` will attempt to set the values
                        # inplace instead of always setting a new array. To retain the old behavior, use either
                        # `df[df.columns[i]] = newvals` or, if columns are non-unique, `df.isetitem(i, newvals)`
                        # FROM -----------------------------------------------------------------------------------------
                        # the_df_return.loc[:, k_field] = \
                        #   the_df_return[k_field].apply(lambda x: nan if isnull(x) else round(x, 2))
                        # the_df_return.loc[:, k_field] = the_df_return[k_field].astype('Float64')
                        # TO -------------------------------------------------------------------------------------------
                        the_df_return[k_field] = \
                            the_df_return[k_field].apply(lambda x: nan if isnull(x) else round(x, 2))
                        the_df_return[k_field] = the_df_return[k_field].astype('Float64')
                        # END ------------------------------------------------------------------------------------------
                    except Exception as an_exception:
                        raise Exception(an_exception, f"Exception raised when typing column "
                                                      f"'{k_field}' into '{v_df_type}'.")

                elif v_df_type == PandasDataframeTyper.str__type__int:
                    try:
                        # JSG as of 22-Jan-2023: FutureWarning ---------------------------------------------------------
                        # FutureWarning: In a future version, `df.iloc[:, i] = newvals` will attempt to set the values
                        # inplace instead of always setting a new array. To retain the old behavior, use either
                        # `df[df.columns[i]] = newvals` or, if columns are non-unique, `df.isetitem(i, newvals)`
                        # FROM -----------------------------------------------------------------------------------------
                        # the_df_return.loc[:, k_field] = \
                        #                             the_df_return[k_field].astype('Int64')
                        # TO -------------------------------------------------------------------------------------------
                        the_df_return[k_field] = \
                            the_df_return[k_field].astype('Int64')
                        # END ------------------------------------------------------------------------------------------
                    except Exception as an_exception:
                        raise Exception(an_exception, f"Exception raised when typing column "
                                                      f"'{k_field}' into '{v_df_type}'.")

                elif v_df_type == PandasDataframeTyper.str__type__str:
                    try:
                        # JSG as of 22-Jan-2023: FutureWarning ---------------------------------------------------------
                        # FutureWarning: In a future version, `df.iloc[:, i] = newvals` will attempt to set the values
                        # inplace instead of always setting a new array. To retain the old behavior, use either
                        # `df[df.columns[i]] = newvals` or, if columns are non-unique, `df.isetitem(i, newvals)`
                        # FROM -----------------------------------------------------------------------------------------
                        # the_df_return.loc[:, k_field] = \
                        #                             the_df_return[k_field].astype('string')
                        # TO -------------------------------------------------------------------------------------------
                        the_df_return[k_field] = \
                            the_df_return[k_field].astype('string')
                        # END ------------------------------------------------------------------------------------------
                    except Exception as an_exception:
                        raise Exception(an_exception, f"Exception raised when typing column "
                                                      f"'{k_field}' into '{v_df_type}'.")

                else:
                    raise Exception(f"We don't know how to handle transco into '{v_df_type}'.")

        PandasDataframeTyper._logger.debug(
            f"Function '{inspect.stack()[0].filename} - {inspect.stack()[0].function}' is returning.")

        return the_df_return



import logging
import os
from inspect import stack
from pathlib import Path

import click
from click import BadParameter


_logger: logging.Logger = logging.getLogger(__name__)


@click.command('dp-demo')
@click.argument('file', type=str)
@click.pass_context
def dp_demo_date_predictability(ctx_context, file):
    """
    Demonstration of how to use commands related to Date Predictability.

    FILE is the path to persist the outputs generated by the Date Predictability commands, either an existing directory
    or an non-existing 'xlsx' file in an existing directory
    """
    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    p_the_file_path: Path = Path(file)
    if not p_the_file_path.is_absolute():
         p_the_file_path = Path(os.getcwd()).joinpath(p_the_file_path)

    # If the file path is an existing directory
    if p_the_file_path.is_dir():
        if p_the_file_path.exists():
            p_the_file_path = p_the_file_path.joinpath("com.enovation.toolbox.predictability.bean.xlsx")
        else:
            raise BadParameter(f"Directory '{p_the_file_path}' does not exist. Please make sure you provide a path to "
                               f"an existing directory.")

    # If the file path is an existing file
    elif p_the_file_path.is_file():
        raise BadParameter(f"File '{p_the_file_path}' does exist. Please make sure you provide a path to a file that "
                           f"does not exist.")

    # If the file path is neither an existing directory, nor an existing file
    else:
        if not p_the_file_path.resolve().parents[0].is_dir():
            raise BadParameter(f"Directory '{p_the_file_path.resolve().parents[0]}' does not exist. Please make sure "
                               f"you provide a path pointing to an existing directory.")
        elif p_the_file_path.suffix != ".xlsx":
            raise BadParameter(f"File '{p_the_file_path.name}' should have extension 'xlsx', not "
                               f"'{p_the_file_path.suffix}'. Please make sure you provide the proper extension.")

    click.echo(f"**************************"
               f"\n* Copy past in your terminal the command generated below, that will do the following:"
               f"\n*\t- Load a demo csv file '{Path(__file__).resolve().parents[0].joinpath('dp_demo.csv')}'"
               f"\n*\t- Rename columns"
               f"\n*\t- Compute date predictability"
               f"\n*\t- Persist the outputs into an excel spreadsheet '{p_the_file_path}'"
               f"\n*\t- Launch a dash server to graph the outputs"
               f"\n**************************")
    click.echo(
        f"enov --verbose "
               
        f"df-load-csv --columns '[\"Project and phase\", \"Projection date\", \"Projected ETA\"]' "
        f"\"{Path(__file__).resolve().parents[0].joinpath('dp_demo.csv')}\" df_raw "
       
        f"df-rename-columns -o df_renamed df_raw "
        "'{\"Project and phase\": \"key\", \"Projection date\": \"date\", "
        "\"Projected ETA\": \"measure\"}' "
        
        f"dp-compute --resample df_renamed df_predictability "
        
        f"dp-persist df_predictability \"{p_the_file_path}\" "

        f"dp-graph df_predictability")

    click.echo(f"**************************"
               f"\n* In case 'enov' is not recognized, copy past the following command..."
               f"\n**************************")
    click.echo(
        f"python3 \"{Path(__file__).resolve().parents[5].joinpath('enov.py')}\" --verbose "
               
        f"df-load-csv --columns '[\"Project and phase\", \"Projection date\", \"Projected ETA\"]' "
        f"\"{Path(__file__).resolve().parents[0].joinpath('dp_demo.csv')}\" df_raw "
       
        f"df-rename-columns -o df_renamed df_raw "
        "'{\"Project and phase\": \"key\", \"Projection date\": \"date\", "
        "\"Projected ETA\": \"measure\"}' "
        
        f"dp-compute --resample df_renamed df_predictability "
        
        f"dp-persist df_predictability \"{p_the_file_path}\" "

        f"dp-graph df_predictability")

    _logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import collections.abc
import copy
import inspect
import json
from inspect import stack
from logging import Logger, getLogger
from pathlib import Path
from typing import Optional

from xlsxwriter.format import Format
from xlsxwriter.utility import xl_cell_to_rowcol
from xlsxwriter.workbook import Workbook
from xlsxwriter.worksheet import Worksheet

from com.enovation.toolbox.excel_dashboard.cell_notation_helper import CellNotationHelper


def copy_then_update(d, u):
    return update(copy.deepcopy(d), u)


def update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = update(d.get(k, {}), v)
        else:
            d[k] = v
    return d


class Widget:
    """
    Default class for widgets.

    A widget depends on 3 components:
    - Python code: logic to print the widget in an excel spreadsheet. eg. print a list of data which evolve over time
    - Json configuration: configuration of the widget for a given application. This configuration is used and reused as
      the application is run. eg. the columns to be printed in the list, along with their formattings
    - Parameters: particular context that is provided at each run of the application. eg. the data to print in the list

    To implement a widget, you have to go through the following steps:
    - Add a dedicated directory that will contain both:
      - The python code of the widget
      - The json schema file to validate widget configuration (same file name, replacing extension '.py' by '.json')
    - The python code of the widget should extend this Widget class
      - Override the function check_parameters: this function validates the parameters provided to the widget. eg. will
        check the configured columns to print in the list effectively exist in the dataframe containing the data to
        print (provided as a DataFrame parameter)
      - Override the function write_to_excel, which is where the logic to print the widget into excel lies.

    To support the above, each widget inherits from the following properties:
    - Class properties:
      - 'dict_config_schema': the json schema to check and validate the widget configuration. This is a class property
        as all instances of the same widget share the same json schema
    - Instance properties:
      - 'str_address': the address (in a sheet) to print the widget
      - 'dict_config': the configuration of the widget (validated against the schema 'dict_config_schema')
      - 'dict_parameters_map': the dictionary to map actual parameters names into the ones expected by the widget
      - 'dict_raw_parameters': the parameters as provided by the user to the widget (before being mapped into the
        parameters expected by the widget)

    TODO:
    - The default formats:
      - _dict_wb_default_format should be associated to the super class Widget, and not to sub_classes cls --> only one
        default across each an every Widget subclasses
      - _dict_ws_default_format should be associated to each and every Widget subclass, getting rid of the widget
        label...
    """

    const_str__class_attribute__dict_config_schema: str = "dict_config_schema"

    # Labels from json schema
    const_str__json_schema__widget_id: str = "$id"

    # Labels from configuration
    const_str__config__widget_id: str = "widget_id"

    # Labels from configuration for the conditional formattings
    const_str__config_label__str_options: str = "options"
    const_str__config_label__str_formula: str = "formula"
    const_str__config_label__str_multi_range: str = "multi_range"
    const_str__config_label__str_range: str = "range"
    const_str__config_label__str_criteria: str = "criteria"
    const_str__config_label__str_type: str = "type"
    const_str__config_label__str_format: str = "format"

    # There is a mechanism to set default formats at various levels in the application:
    # - Level 1: the entire workbook can have a default format
    # - Level 2: each worksheet can have their own default formats
    # - Level 3: each widget can have themselves their own default formats
    # Whenever default formats exist at various levels, the mechanism has the logic to consolidate.
    #
    # Illustration:
    # - The workbook has a default format level 1: wrap, color black
    # - The sheet has a default format level 2: underline
    # - The widget has a default format level 3: italic, color green
    # - The function get_format is called for a format: no italic
    # - The resulting format returned is: wrap, color green, underline, no italic
    #
    # The following properties and functions support this logic:
    # - _dict_wb_default_format: class property that contains the unique default format that is set at the level of
    #   the workbook
    # - _dict_ws_default_formats_raw: class property that contains the various default formats by worksheet. The
    #   worksheet label is used as the key. This is the raw format as configured, not inheriting from the default
    #   format level 1 set at the workbook level. The getter method dict_ws_default_formats will return the sheet
    #   clean format, inheriting from the default format level 1 set at the workbook level
    # - _dict_default_format_raw: instance property that contains the widget default format. This is the raw format as
    #   configured, not inheriting from the default formats level 1 or 2 set at the workbook or worksheet levels. The
    #   getter method dict_default_format will return the widget clean format, inheriting from both the default format
    #   levels 1 and 2, set at the workbook and worksheet levels
    _dict_wb_default_format: Optional[dict] = None

    @classmethod
    def set_dict_wb_default_format(cls, dict_default_format: dict):
        cls._dict_wb_default_format = dict_default_format

    @classmethod
    def get_dict_wb_default_format(cls) -> dict:
        if cls._dict_wb_default_format is None:
            getLogger(__name__).error(
                f"The class property _dict_wb_default_format was not initialized, which is not expected. "
                f"Ensure you call function set_dict_wb_default_format."
            )
            cls._dict_wb_default_format = {}
            # raise Exception("The class property _dict_wb_default_format was not initialized, which is not expected. "
            #                 "Ensure you call function set_dict_wb_default_format.")
        return cls._dict_wb_default_format

    _dict_ws_default_formats: dict = {}

    @classmethod
    def add_dict_ws_default_format(cls, str_sheet_name: str, dict_default_format: dict):
        if str_sheet_name in cls._dict_ws_default_formats:
            raise Exception(f"The class property _dict_ws_default_formats already has a default format for the sheet "
                            f"named '{str_sheet_name}', which is not expected. "
                            f"Ensure you call function add_dict_ws_default_format only once.")
        _dict_the_format: dict = cls.get_dict_wb_default_format()
        _dict_the_format = copy_then_update(_dict_the_format, dict_default_format)
        cls._dict_ws_default_formats[str_sheet_name] = _dict_the_format

    @classmethod
    def get_dict_ws_default_format(cls, str_sheet_name: str) -> dict:
        if str_sheet_name not in cls._dict_ws_default_formats:
            getLogger(__name__).error(
                f"The class property _dict_ws_default_formats does not have default format for the sheet "
                f"named '{str_sheet_name}', which is not expected. "
                f"Ensure you call function add_dict_ws_default_format, even with an empty dictionary."
            )
            cls.add_dict_ws_default_format(str_sheet_name=str_sheet_name, dict_default_format={})
        return cls._dict_ws_default_formats[str_sheet_name]

    def _get_dict_default_format(self, dict_default_format: dict, str_sheet_name: str) -> dict:
        # if self._dict_wb_default_format is None:
        #     self._logger.error(
        #         f"The class property _dict_wb_default_format was not initialized, which is not expected. "
        #         f"Ensure you call function set_dict_wb_default_format."
        #     )
        #     self._dict_wb_default_format = {}
        # if str_sheet_name not in self._dict_ws_default_formats:
        #     self._logger.error(
        #         f"The class property _dict_ws_default_formats already has a default format for the sheet "
        #         f"named '{str_sheet_name}', which is not expected. "
        #         f"Ensure you call function add_dict_ws_default_format only once."
        #     )
        #     self.add_dict_ws_default_format(str_sheet_name=str_sheet_name, dict_default_format={})

        _dict_the_format = self.get_dict_ws_default_format(str_sheet_name=str_sheet_name)
        _dict_the_format = copy_then_update(_dict_the_format, dict_default_format)
        return _dict_the_format

    @classmethod
    def reinitialize_default_formats(cls):
        cls._dict_wb_default_format = None
        cls._dict_ws_default_formats = {}

    _logger: Logger = getLogger(__name__)

    # the address (in a sheet) to print the widget
    @property
    def str_address(self) -> str:
        return self._str_address

    # the address (in a sheet) to print the widget as a tuple(row, col)
    @property
    def tpl_address(self) -> tuple[int, int]:
        return xl_cell_to_rowcol(self.str_address)

    # the map from actual parameters to the ones expected by the widget (only expected when these are different)
    # - key=expected parameter name, as the widget code will refer to
    # - value=actual parameter name, as provided by the wider application.
    @property
    def dict_parameters_map(self) -> dict:
        return self._dict_parameters_map

    # the configuration of the widget (validated against the schema 'dict_config_schema')
    @property
    def dict_config(self) -> dict:
        return self._dict_config

    # the raw parameters provided by the user, before they are renamed and checked
    @property
    def dict_raw_parameters(self) -> dict:
        return self._dict_raw_parameters

    # the default widget format, inheriting from the sheet and the workbook default formats
    @property
    def dict_default_format(self) -> dict:
        return self._dict_default_format

    @property
    def lst_conditional_formattings(self) -> list:
        return self._lst_conditional_formattings

    def __init__(
            self,
            str_address: str,
            dict_parameters_map: dict,
            dict_config: dict,
            dict_raw_parameters: dict,
            str_sheet_name: str,
            dict_default_format: dict
    ):
        """
        Initialize the widget:
        - Step 1, we set the widget properties:
          - from nodes in the json configuration: address, parameters map, config
          - from parameters: raw parameters
        - Step 2, we process the parameters:
          - Rename parameters to fit the labels as expected by the widget
          - Check parameters to ensure their values are correct, and set widget properties, to use these parameters
            while printing into the Excel spreadsheet --> function 'check_parameters_and_set_properties' to be
            overridden by each widget

        :param str_address: the address in a sheet to print the widget
        :param dict_config: the widget configuration (aka the "config" node in the json)
        :param dict_parameters_map: the dictionary to map users parameters to the ones expected by the widget:
                                    - key: the expected label
                                    - value: the actual label
        :param dict_raw_parameters: the parameters as labelled and provided by the users
        :param str_sheet_name: the name of the sheet in which the widget is instantiated (required to handle default
            formats inheritance, from the workbook to the worksheet to the widget)
        :param dict_default_format: the default format set at the level of the widget. If such default format is not
            configured, we expect an empty dictionary
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # Step 1, we set the widget properties: address, config, parameters map, raw parameters
        self._str_address = str_address
        self._dict_config = copy.deepcopy(dict_config)  # Required, as we might transform some format in the config
        self._dict_parameters_map = dict_parameters_map
        self._dict_raw_parameters = dict_raw_parameters

        # Conditional formattings will be set through the dedicated function set_conditional_formattings, which is
        # called through the function check parameters and set properties
        self._lst_conditional_formattings: list[dict] = []

        # Step 1b, we instantiate the default format for the widget, inheriting from default formats set at the level
        # of the workbook and the worksheets
        self._dict_default_format = self._get_dict_default_format(
            dict_default_format=dict_default_format,
            str_sheet_name=str_sheet_name
        )

        # Step 2, we rename, check and get the widget parameters
        self._rename_check_and_set_properties(
            dict_parameters_map=self._dict_parameters_map,
            dict_config=self._dict_config,
            dict_raw_parameters=self._dict_raw_parameters
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def _rename_check_and_set_properties(
            self,
            dict_parameters_map: dict,
            dict_config: dict,
            dict_raw_parameters: dict,
    ):
        """
        Function that processes the parameters:
        - Rename parameters to fit the labels as expected by the widget
        - Check parameters to ensure their values are correct, and set the widget properties to be used
          while printing into the Excel spreadsheet --> function 'check_parameters_and_get_parameters' to be
          overridden by each widget.

        While renaming parameters, if an actual parameter name to map is missing, then an exception is raised.

        :param dict_parameters_map: dictionary to map the actual parameters into the ones expected by the widget
                                    - key: expected parameter name, as referenced within the widget
                                    - value: actual parameter name, as provided as part of the wider application
        :param dict_config: dictionary with the widget configuration
        :param dict_raw_parameters: dictionary that contains the raw parameters to rename into the ones expected by the
                                    widget
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We initialize the dictionary to update the parameters
        dict_renamed_parameters: dict = {}

        # Step 1, we rename the parameters, using the parameters map
        for k_expected, v_actual in dict_parameters_map.items():
            if v_actual in dict_raw_parameters:
                dict_renamed_parameters[k_expected] = dict_raw_parameters[v_actual]

            else:
                raise Exception(f"When trying to map parameter '{k_expected}' expected by the widget, the actual "
                                f"parameter '{v_actual}' could not be found.")

        # We duplicate the dictionary with all parameters, that we update with renamed parameters
        dict_raw_and_renamed_parameters: dict = dict(dict_raw_parameters)
        dict_raw_and_renamed_parameters.update(dict_renamed_parameters)

        # We eventually call the function 'check_parameters_and_get_parameters' that is to be overridden by each widget
        self.check_parameters_and_set_properties(
            dict_parameters=dict_raw_and_renamed_parameters,
            dict_config=dict_config,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def check_parameters_and_set_properties(
            self,
            dict_parameters: dict,
            dict_config: dict,
    ):
        """
        Function that checks the parameters that are required by the widget:
        - If all expected parameters are correct, the widget properties are set and the function returns (no value
          returned)
        - Otherwise, an exception is thrown.

        Checking parameters means:
        - Check the parameter exists, in case it is a mandatory one. eg. mandatory array of data to print in a list
        - Check the format is as expected. eg. has to be a DataFrame
        - Check the values are correct. eg. numeric vs alpha, etc.
        - Check coherence with config. eg. columns configured as "to be printed" effectively exist in the DataFrame.

        Setting properties means:
        - Set the widget instance properties from the parameters/ config
        - So these properties can later be used while writing into Excel.

        Properties for widget_cover_sheet are:
        - str_narrative
        - lst_tokens

        :param dict_parameters: the parameters provided to the module
        :param dict_config: the configuration of the module
        """
        raise Exception(f"Function to be overridden...")

    def get_format(
            self,
            dict_format: dict,
            wb_workbook: Workbook
    ) -> Format:
        """
        Function that define a format object, from its properties.

        :param dict_format:
        :param wb_workbook:
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _dict_the_format: dict = copy_then_update(
            self.dict_default_format,
            dict_format
        )

        _format: Format = wb_workbook.add_format(properties=_dict_the_format)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
        return _format

    def set_conditional_formattings(
            self,
            lst_conditional_formattings: list[dict],
            str_reference_address: str,
            str_column_range_limit: str = None,
            str_row_range_limit: str = None
    ):
        """
        Function that processes the conditional formattings configured, and transform the following attributes:
        - The "range"
        - The "options"."multi_range", if it exists
        - The "options"."criteria" for the conditional formatting "options"."type"=="formula"

        This function is called from the sub-widget classes, when the function check_parameters_and_set_properties is
        called
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        for i_conditional_formatting in lst_conditional_formattings:

            _dict_i_conditional_formatting: dict = copy.deepcopy(i_conditional_formatting)

            # We process the range
            _dict_i_conditional_formatting[Widget.const_str__config_label__str_range] = \
                Widget._process_conditional_formatting__range(
                    str_range=i_conditional_formatting[Widget.const_str__config_label__str_range],
                    str_reference_address=str_reference_address,
                    str_column_range_limit=str_column_range_limit,
                    str_row_range_limit=str_row_range_limit
                )

            # We process the multi-range, if any
            if Widget.const_str__config_label__str_multi_range in \
                    _dict_i_conditional_formatting[Widget.const_str__config_label__str_options]:
                _dict_i_conditional_formatting[
                    Widget.const_str__config_label__str_options][
                    Widget.const_str__config_label__str_multi_range] = \
                    Widget._process_conditional_formatting__multi_range(
                        str_multi_range=i_conditional_formatting[
                            Widget.const_str__config_label__str_options][
                            Widget.const_str__config_label__str_multi_range],
                        str_reference_address=str_reference_address,
                    str_column_range_limit=str_column_range_limit,
                    str_row_range_limit=str_row_range_limit
                    )

            # If the conditional formatting contains a formula
            if _dict_i_conditional_formatting[
                Widget.const_str__config_label__str_options][
                Widget.const_str__config_label__str_type] == \
                    Widget.const_str__config_label__str_formula:
                _dict_i_conditional_formatting[
                    Widget.const_str__config_label__str_options][
                    Widget.const_str__config_label__str_criteria] = \
                    Widget._process_conditional_formatting__formula(
                        str_formula=i_conditional_formatting[
                            Widget.const_str__config_label__str_options][
                            Widget.const_str__config_label__str_criteria],
                        str_reference_address=str_reference_address
                    )

            self._lst_conditional_formattings.append(_dict_i_conditional_formatting)

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    @staticmethod
    def _process_conditional_formatting__range(
            str_range: str,
            str_reference_address: str,
            str_column_range_limit: str = None,
            str_row_range_limit: str = None
    ) -> str:
        """
        Function that processes a range for a given conditional formatting. It receives a range or cell address, and
        seeks to:
        - Translate it relative to a reference address,
        - Transform it, if expressed as a column or a row, into a range.

        The limits are set relatively to the widget, and have to be translated to the reference cell

        An exception is thrown in case:
        - The input range is not pattern cell, range or column
        - The input range is a column or a row range, but there is not limit set (to transform into a full range)
        - There is any dollar in the range provided
        - The outputs range is not correct cell or range.

        A warning is logged in case there is any limit provided but not used.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called with parameters:"
                                  f"\n\t- str_range: '{str_range}'"
                                  f"\n\t- str_reference_address: '{str_reference_address}'"
                                  f"\n\t- str_column_range_limit: '{str_column_range_limit}'"
                                  f"\n\t- str_row_range_limit: '{str_row_range_limit}'")

        _str_the_return: str = str_range

        # If we do not have a proper pattern for the input str_range, we raise an exception
        _str_the_pattern: str = CellNotationHelper.get_address_pattern(str_address=str_range)
        if (
                CellNotationHelper.is_pattern_cell(str_pattern=_str_the_pattern)
                | CellNotationHelper.is_pattern_range(str_pattern=_str_the_pattern)
                | CellNotationHelper.is_pattern_column(str_pattern=_str_the_pattern)
                | CellNotationHelper.is_pattern_row(str_pattern=_str_the_pattern)
        ) is False:
            raise Exception(f"The parameter str_range '{str_range}' has an incorrect pattern '{_str_the_pattern}'. "
                            f"It should be either a cell, a range or a column.")

        # If there is any dollar in the range provided
        if "$" in str_range:
            raise Exception(f"There is '$' in the parameter str_range='{str_range}', which is not expected.")

        # If we do receive column or row ranges, we need the associated limits
        if CellNotationHelper.is_pattern_column(str_pattern=_str_the_pattern):
            if str_row_range_limit is None:
                raise Exception(f"The parameter str_range '{str_range}' is a column range, but no str_row_range_limit "
                                f"is provided to transform it into a range... Which is not expected.")
            if str_column_range_limit:
                getLogger(__name__).warning(
                    f"The parameter str_range '{str_range}' is a column range, but one useless str_column_range_limit "
                    f"is provided... Only str_row_range_limit is required to transform it into a range... Maybe "
                    f"a configuration inconsistency?")

            # We override the column range with a range
            _str_the_return = CellNotationHelper.transform_columns_intersect_rows_into_range_or_cell(
                str_columns=_str_the_return,
                str_rows=CellNotationHelper.transform_address_into_relative_and_absolute_address(
                    str_address=str_row_range_limit,
                    b_first_row_at=True,
                    b_last_row_at=True
                ),
                b_check_correct=False
            )

        elif CellNotationHelper.is_pattern_row(str_pattern=_str_the_pattern):
            if str_column_range_limit is None:
                raise Exception(f"The parameter str_range '{str_range}' is a row range, but no str_column_range_limit "
                                f"is provided to transform it into a range... Which is not expected.")
            if str_row_range_limit:
                getLogger(__name__).warning(
                    f"The parameter str_range '{str_range}' is a row range, but one useless str_row_range_limit "
                    f"is provided... Only str_column_range_limit is required to transform it into a range... Maybe "
                    f"a configuration inconsistency?")

            # We override the row range with a range
            _str_the_return = CellNotationHelper.transform_columns_intersect_rows_into_range_or_cell(
                str_columns=CellNotationHelper.transform_address_into_relative_and_absolute_address(
                    str_address=str_column_range_limit,
                    b_first_col_at=True,
                    b_last_col_at=True
                ),
                str_rows=_str_the_return,
                b_check_correct=False
            )

        # We eventually translate the range vis-a-vis the reference
        _str_the_return = CellNotationHelper.translate_address_to_reference(
            str_address=_str_the_return,
            str_reference_cell=str_reference_address
        )

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning "
                                  f"'{_str_the_return}'.")
        return _str_the_return

    @staticmethod
    def _process_conditional_formatting__multi_range(
            str_multi_range: str,
            str_reference_address: str,
            str_column_range_limit: str = None,
            str_row_range_limit: str = None
    ) -> str:
        """
        Function that processes a multi-range for a given conditional formatting. Multi-range is configured as
        multiple range tokens that separated by space.

        The function will first tokenize the multi-range into range token, that can then be processed through the
        function _process_conditional_formatting__range.

        The function _process_conditional_formatting__range will:
        - Translate each range token relative to a reference address,
        - Transform each range token, if expressed as a column, into a range.

        Eventually, we concatenate the processed range tokens back into one single multi-range

        An exception is thrown in case:
        - The input multi-range has one range token:
          - Which is not pattern cell, range or column
          - Which is a column or a row range, but there is not limit set (to transform into a full range)
          - Which has any dollar
        - The outputs multi-range is not correct concatenation of cells or ranges.

        A warning is logged in case there is any limit provided but not used.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called with parameters:"
                                  f"\n\t- str_multi_range: '{str_multi_range}'"
                                  f"\n\t- str_reference_address: '{str_reference_address}'"
                                  f"\n\t- str_column_range_limit: '{str_column_range_limit}'"
                                  f"\n\t- str_row_range_limit: '{str_row_range_limit}'")

        _lst_str_range_token: list[str] = str_multi_range.split()
        _lst_the_return: list[str] = []
        _str_the_return: str

        for i_range_token in _lst_str_range_token:
            _lst_the_return.append(
                Widget._process_conditional_formatting__range(
                    str_range=i_range_token,
                    str_reference_address=str_reference_address,
                    str_column_range_limit=str_column_range_limit,
                    str_row_range_limit=str_row_range_limit
                )
            )

        _str_the_return = " ".join(_lst_the_return)

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning "
                                  f"'{_str_the_return}'.")
        return _str_the_return

    @staticmethod
    def _process_conditional_formatting__formula(
            str_formula: str,
            str_reference_address: str
    ) -> str:
        """
        Function that processes a formula for a given conditional formatting. It receives a formula, and
        seeks to:
        - Translate the relative addresses to a reference, if any,
        - Transform them.

        Illustration: for reference address B3
        - @A would become: B
        - @A1 would become: B1
        - A@1 would become: A3
        - @A:@A would become: B:B
        - @1:@1 would become: 3:3

        An exception is thrown in case:
        - The formula cannot be processed
        - Any error in the addresses to transform.
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called with parameters:"
                                  f"\n\t- str_formula: '{str_formula}'"
                                  f"\n\t- str_reference_address: '{str_reference_address}'")

        _str_the_return: str = CellNotationHelper.translate_formula_to_reference(
            str_reference_address=str_reference_address,
            str_formula=str_formula
        )

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning "
                                  f"'{_str_the_return}'.")
        return _str_the_return

    def write_to_excel(
            self,
            wb_workbook: Workbook,
            ws_worksheet: Worksheet
    ):
        """
        Function that writes the widget into the excel spreadsheet. When called, the widget is fully initialized,
        with all its properties set.

        Note: it is while writing into excel that we
        - Transcode formats, from "format config" as dictionary to "format" object as expected by xlsxwriter, by
          calling the function get_format
        - Write the conditional formattings by calling the function the function write_conditional_formattings.

        :param ws_worksheet:
        :param wb_workbook:
        """
        raise Exception(f"Function to be overridden...")

    def write_conditional_formattings(
            self,
            wb_workbook: Workbook,
            ws_worksheet: Worksheet
    ):

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called .")

        # We add the conditional formattings
        for i_cond_format in self.lst_conditional_formattings:
            _str_the_range: str = i_cond_format[self.const_str__config_label__str_range]

            _dict_the_options: dict = i_cond_format[self.const_str__config_label__str_options]

            _dict_the_format: dict = _dict_the_options[self.const_str__config_label__str_format]

            _obj_the_format: Format = self.get_format(
                dict_format=_dict_the_format,
                wb_workbook=wb_workbook
            )

            _dict_the_options[self.const_str__config_label__str_format] = _obj_the_format

            ws_worksheet.conditional_format(
                _str_the_range,
                options=_dict_the_options
            )
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning .")

    @classmethod
    def get_dict_config_schema(
            cls,
    ) -> dict:
        """
        Function that returns the json schema to configure the widget. This json schema should comply to the following
        rules:
        - Should be labelled the same way as the python file
        - Except ".py" extension is "json".

        In case such a json does not exist, the function returns an empty dictionary.

        :return: the json schema as a dictionary to validate the configuration of the widget
        """
        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # We get the class attribute "dict_config_schema"
        dict_the_return: dict = getattr(cls, Widget.const_str__class_attribute__dict_config_schema, None)

        # If the class attribute "dict_config_schema was not yet set
        if dict_the_return is None:

            # We get the file which defines the class
            _path_to_python_file: Path = Path(inspect.getfile(cls))
            _path_to_json_file: Path = _path_to_python_file.with_suffix('.json')
            if _path_to_json_file.is_file():
                try:
                    with _path_to_json_file.open() as _json_schema:
                        dict_the_return = json.load(_json_schema)
                except Exception as _exception:
                    raise Exception(f"Could not load json schema '{_json_schema}'.") from _exception
            else:
                raise Exception(f"When processing widget '{_path_to_python_file}', we could not find the "
                                f"associated json schema, expected under '{_path_to_json_file}'. This is not "
                                f"expected, and the process had to quit...")

            # We persist the class attribute for other instances of the widget
            setattr(cls, Widget.const_str__class_attribute__dict_config_schema, dict_the_return)

        getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return dict_the_return

    @classmethod
    def get_widget_id(
            cls,
    ) -> str:
        if cls.const_str__json_schema__widget_id not in cls.get_dict_config_schema():
            raise Exception(f"When processing widget '{cls}', could not find node "
                            f"'{cls.const_str__json_schema__widget_id}' in the json schema, which is not expected.")

        return cls.get_dict_config_schema()[cls.const_str__json_schema__widget_id]



from logging import Logger, getLogger
from inspect import stack
from pandas import DataFrame
from openpyxl import load_workbook, Workbook

from com.enovation.helper.pandas_dataframe_typer import PandasDataframeTyper


class ExcelLoaderBean:
    _dict_tables: dict

    def __init__(
            self,
            dict_tables: dict,
    ):
        self._dict_tables = dict_tables

    @property
    def dict_tables(self) -> dict:
        return self._dict_tables


class ExcelLoader:
    _logger: Logger = getLogger(__name__)

    def load_tables(
            self,
            str_path: str,
            dict_tables_columns_types: dict = None
    ) -> ExcelLoaderBean:

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called with parameters:"
                           f"\n\t- str_path: {str_path}")

        # read file
        obj_the_workbook: Workbook = load_workbook(filename=str_path, data_only=True)

        _dict_the_tables: dict = {}

        # For each sheet
        for i_sheet_name in obj_the_workbook.sheetnames:

            # We get the sheet
            obj_the_worksheet = obj_the_workbook[i_sheet_name]

            # For each table in this sheet
            for k_table_name, k_table_range in obj_the_worksheet.tables.items():

                # parse the data within the ref boundary
                obj_the_data = obj_the_worksheet[k_table_range]

                # extract the data
                # the inner list comprehension gets the values for each cell in the table
                lst_the_content: list = \
                    [
                        [
                            i_cell.value for i_cell in i_line
                        ]
                        for i_line in obj_the_data
                    ]

                lst_the_header: list = lst_the_content[0]

                # the contents ... excluding the header
                lst_the_lines: list = lst_the_content[1:]

                # create dataframe with the column names
                # and pair table name with dataframe
                df_the_dataframe = DataFrame(lst_the_lines, columns=lst_the_header)

                # We retype in case we have a dictionary
                if dict_tables_columns_types is not None:
                    if k_table_name in dict_tables_columns_types:
                        df_the_dataframe = PandasDataframeTyper.type(
                            df_to_type=df_the_dataframe,
                            dict_columns_to_type=dict_tables_columns_types[k_table_name]
                        )

                _dict_the_tables[k_table_name] = df_the_dataframe

        obj_the_return: ExcelLoaderBean = ExcelLoaderBean(dict_tables=_dict_the_tables)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

        return obj_the_return
